import unittest2 as unittest

import mock

import datetime
import timex

from winchester import debugging
from winchester import db as winch_db
from winchester import pipeline_manager
from winchester.models import StreamState


class TestPipeline(unittest.TestCase):
    def setUp(self):
        super(TestPipeline, self).setUp()
        self.debugger = debugging.NoOpDebugger()
        self.fake_stream = mock.MagicMock(name="fake_stream")
        self.fake_stream.id = "stream-1234"

    def test_check_handler_config(self):

        handler_map = {'test_thing': "blah"}
        c = pipeline_manager.Pipeline.check_handler_config("test_thing", handler_map)
        self.assertIsInstance(c, dict)
        self.assertIn('name', c)
        self.assertIn('params', c)
        self.assertIsInstance(c['params'], dict)
        self.assertEqual(c['name'], 'test_thing')
        self.assertEqual(c['params'], {})

        conf = dict(name='test_thing')
        c = pipeline_manager.Pipeline.check_handler_config(conf, handler_map)
        self.assertIsInstance(c, dict)
        self.assertIn('name', c)
        self.assertIn('params', c)
        self.assertIsInstance(c['params'], dict)
        self.assertEqual(c['name'], 'test_thing')
        self.assertEqual(c['params'], {})

        conf = dict(name='test_thing', params={'book': 42})
        c = pipeline_manager.Pipeline.check_handler_config(conf, handler_map)
        self.assertIsInstance(c, dict)
        self.assertIn('name', c)
        self.assertIn('params', c)
        self.assertIsInstance(c['params'], dict)
        self.assertEqual(c['name'], 'test_thing')
        self.assertEqual(c['params'], {'book': 42})

        with self.assertRaises(pipeline_manager.PipelineConfigError):
            c = pipeline_manager.Pipeline.check_handler_config("other_thing", handler_map)

        with self.assertRaises(pipeline_manager.PipelineConfigError):
            conf = dict(params={'book': 42})
            c = pipeline_manager.Pipeline.check_handler_config(conf, handler_map)

    def test_init(self):
        conf = [dict(name='test_thing', params={'book': 42})]
        handler_class = mock.MagicMock()
        handler_map = {'test_thing':  handler_class}
        p = pipeline_manager.Pipeline("test_pipeline", conf, handler_map)
        self.assertEqual(p.name, "test_pipeline")
        self.assertEqual(len(p.handlers), 1)
        self.assertIs(handler_class.return_value, p.handlers[0])
        handler_class.assert_called_once_with(book=42)

    def test_handle_events(self):
        test_events = [dict(message_id="t000-0001"),
                       dict(message_id="t000-0002"),
                       dict(message_id="t000-0003")]
        new_events = [dict(message_id="t000-0004")]
        conf = [dict(name='test_thing', params={}),
                dict(name='other_thing', params={}),
                dict(name='some_thing', params={})]
        handler_class1 = mock.MagicMock(name='handler1')
        handler_class2 = mock.MagicMock(name='handler2')
        handler_class3 = mock.MagicMock(name='handler3')
        handler_class3.return_value.handle_events.return_value = test_events + new_events

        handler_map = {'test_thing':  handler_class1,
                       'other_thing': handler_class2,
                       'some_thing':  handler_class3}
        p = pipeline_manager.Pipeline("test_pipeline", conf, handler_map)
        p.commit = mock.MagicMock(name='commit')
        p.rollback = mock.MagicMock(name='rollback')

        ret = p.handle_events(test_events, self.fake_stream, self.debugger)
        handler_class1.return_value.handle_events.assert_called_once_with(test_events, p.env)
        events1 = handler_class1.return_value.handle_events.return_value
        handler_class2.return_value.handle_events.assert_called_once_with(events1, p.env)
        events2 = handler_class2.return_value.handle_events.return_value
        handler_class3.return_value.handle_events.assert_called_once_with(events2, p.env)
        p.commit.assert_called_once_with(self.debugger)
        self.assertFalse(p.rollback.called)
        self.assertEqual(ret, new_events)

    def test_handle_events_error(self):
        test_events = [dict(message_id="t000-0001"),
                       dict(message_id="t000-0002"),
                       dict(message_id="t000-0003")]
        conf = [dict(name='test_thing', params={}),
                dict(name='other_thing', params={}),
                dict(name='some_thing', params={})]
        handler_class1 = mock.MagicMock(name='handler1')
        handler_class2 = mock.MagicMock(name='handler2')
        handler_class3 = mock.MagicMock(name='handler3')

        class WhackyError(Exception):
            pass

        handler_class2.return_value.handle_events.side_effect = WhackyError("whoa!")

        handler_map = {'test_thing':  handler_class1,
                       'other_thing': handler_class2,
                       'some_thing':  handler_class3}
        p = pipeline_manager.Pipeline("test_pipeline", conf, handler_map)
        p.commit = mock.MagicMock(name='commit')
        p.rollback = mock.MagicMock(name='rollback')

        with self.assertRaises(pipeline_manager.PipelineExecutionError):
            p.handle_events(test_events, self.fake_stream, self.debugger)
        p.rollback.assert_called_once_with(self.debugger)
        self.assertFalse(p.commit.called)

    def test_commit(self):
        conf = [dict(name='test_thing', params={}),
                dict(name='other_thing', params={}),
                dict(name='some_thing', params={})]
        handler_class1 = mock.MagicMock(name='handler1')
        handler_class2 = mock.MagicMock(name='handler2')
        handler_class3 = mock.MagicMock(name='handler3')

        handler_map = {'test_thing':  handler_class1,
                       'other_thing': handler_class2,
                       'some_thing':  handler_class3}
        p = pipeline_manager.Pipeline("test_pipeline", conf, handler_map)
        p.commit(self.debugger)
        handler_class1.return_value.commit.assert_called_once_with()
        handler_class2.return_value.commit.assert_called_once_with()
        handler_class3.return_value.commit.assert_called_once_with()

    def test_commit_with_error(self):
        conf = [dict(name='test_thing', params={}),
                dict(name='other_thing', params={}),
                dict(name='some_thing', params={})]
        handler_class1 = mock.MagicMock(name='handler1')
        handler_class2 = mock.MagicMock(name='handler2')
        handler_class3 = mock.MagicMock(name='handler3')

        class WhackyError(Exception):
            pass

        handler_class2.return_value.commit.side_effect = WhackyError("whoa!")

        handler_map = {'test_thing':  handler_class1,
                       'other_thing': handler_class2,
                       'some_thing':  handler_class3}
        p = pipeline_manager.Pipeline("test_pipeline", conf, handler_map)
        p.commit(self.debugger)
        handler_class1.return_value.commit.assert_called_once_with()
        handler_class2.return_value.commit.assert_called_once_with()
        handler_class3.return_value.commit.assert_called_once_with()

    def test_rollback(self):
        conf = [dict(name='test_thing', params={}),
                dict(name='other_thing', params={}),
                dict(name='some_thing', params={})]
        handler_class1 = mock.MagicMock(name='handler1')
        handler_class2 = mock.MagicMock(name='handler2')
        handler_class3 = mock.MagicMock(name='handler3')

        handler_map = {'test_thing':  handler_class1,
                       'other_thing': handler_class2,
                       'some_thing':  handler_class3}
        p = pipeline_manager.Pipeline("test_pipeline", conf, handler_map)
        p.rollback(self.debugger)
        handler_class1.return_value.rollback.assert_called_once_with()
        handler_class2.return_value.rollback.assert_called_once_with()
        handler_class3.return_value.rollback.assert_called_once_with()

    def test_rollback_with_error(self):
        conf = [dict(name='test_thing', params={}),
                dict(name='other_thing', params={}),
                dict(name='some_thing', params={})]
        handler_class1 = mock.MagicMock(name='handler1')
        handler_class2 = mock.MagicMock(name='handler2')
        handler_class3 = mock.MagicMock(name='handler3')

        class WhackyError(Exception):
            pass

        handler_class2.return_value.rollback.side_effect = WhackyError("whoa!")

        handler_map = {'test_thing':  handler_class1,
                       'other_thing': handler_class2,
                       'some_thing':  handler_class3}
        p = pipeline_manager.Pipeline("test_pipeline", conf, handler_map)
        p.rollback(self.debugger)
        handler_class1.return_value.rollback.assert_called_once_with()
        handler_class2.return_value.rollback.assert_called_once_with()
        handler_class3.return_value.rollback.assert_called_once_with()


class TestPipelineManager(unittest.TestCase):

    def setUp(self):
        super(TestPipelineManager, self).setUp()
        self.debugger = debugging.NoOpDebugger()

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_complete_stream_nopurge(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        pm.db = mock.MagicMock(spec=pm.db)
        pm.purge_completed_streams = False
        stream = "test stream"
        pm._complete_stream(stream)
        pm.db.set_stream_state.assert_called_once_with(stream, StreamState.completed)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_complete_stream_purge(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        pm.db = mock.MagicMock(spec=pm.db)
        pm.purge_completed_streams = True
        stream = "test stream"
        pm._complete_stream(stream)
        pm.db.purge_stream.assert_called_once_with(stream)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_error_stream(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        pm.db = mock.MagicMock(spec=pm.db)
        stream = "test stream"
        pm._error_stream(stream)
        pm.db.set_stream_state.assert_called_once_with(stream, StreamState.error)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_expire_error_stream(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        pm.db = mock.MagicMock(spec=pm.db)
        stream = "test stream"
        pm._expire_error_stream(stream)
        pm.db.set_stream_state.assert_called_once_with(stream, StreamState.expire_error)

    @mock.patch('winchester.pipeline_manager.Pipeline', autospec=True)
    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_run_pipeline(self, mock_config_wrap, mock_pipeline):
        pm = pipeline_manager.PipelineManager('test')
        pm.db = mock.MagicMock(spec=pm.db, name='db')
        trigger_def = mock.MagicMock(name='trigger_def')
        trigger_def.debugger = self.debugger
        pipeline_name = "test"
        pipeline_config = mock.MagicMock(name='pipeline_config')
        stream = mock.MagicMock(name='stream')
        stream.name = "test"
        pm.add_new_events = mock.MagicMock(name='add_new_events')
        pm.pipeline_handlers = mock.MagicMock(name='pipeline_handlers')

        ret = pm._run_pipeline(stream, trigger_def, pipeline_name,
                               pipeline_config)
        pm.db.get_stream_events.assert_called_once_with(stream)
        mock_pipeline.assert_called_once_with(pipeline_name, pipeline_config,
                                              pm.pipeline_handlers)

        pipeline = mock_pipeline.return_value
        pipeline.handle_events.assert_called_once_with(
            pm.db.get_stream_events.return_value, stream, self.debugger)
        pm.add_new_events.assert_called_once_with(
            mock_pipeline.return_value.handle_events.return_value)
        self.assertTrue(ret)

    @mock.patch('winchester.pipeline_manager.Pipeline', autospec=True)
    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_run_pipeline_with_error(self, mock_config_wrap, mock_pipeline):
        pm = pipeline_manager.PipelineManager('test')
        pm.db = mock.MagicMock(spec=pm.db, name='db')
        trigger_def = mock.MagicMock(name='trigger_def')
        trigger_def.debugger = self.debugger
        pipeline_name = "test"
        pipeline_config = mock.MagicMock(name='pipeline_config')
        stream = mock.MagicMock(name='stream')
        stream.name = "test"
        pm.add_new_events = mock.MagicMock(name='add_nemw_events')
        pm.pipeline_handlers = mock.MagicMock(name='pipeline_handlers')
        pipeline = mock_pipeline.return_value
        pipeline.handle_events.side_effect = \
                    pipeline_manager.PipelineExecutionError('test', 'thing')

        ret = pm._run_pipeline(stream, trigger_def, pipeline_name,
                               pipeline_config)

        pm.db.get_stream_events.assert_called_once_with(stream)
        mock_pipeline.assert_called_once_with(pipeline_name, pipeline_config,
                                              pm.pipeline_handlers)

        pipeline.handle_events.assert_called_once_with(
            pm.db.get_stream_events.return_value, stream, self.debugger)
        self.assertFalse(pm.add_new_events.called)
        self.assertFalse(ret)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_fire_stream(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        stream = mock.MagicMock(name='stream')
        stream.name = 'test'
        pm.db = mock.MagicMock(spec=pm.db, name='db')
        pm.db.set_stream_state.return_value = stream
        trigger_def = mock.MagicMock(name='trigger_def')
        trigger_def.fire_pipeline = 'test_fire_pipeline'
        pm.trigger_map = dict(test=trigger_def)
        pipeline_config = mock.MagicMock(name='pipeline_config')
        pm.pipeline_config = dict(test_fire_pipeline=pipeline_config)
        pm._error_stream = mock.MagicMock(name='_error_stream')
        pm._complete_stream = mock.MagicMock(name='_complete_stream')
        pm._run_pipeline = mock.MagicMock(name='_run_pipeline')
        pm._run_pipeline.return_value = True

        ret = pm.fire_stream(stream)
        pm.db.set_stream_state.assert_called_once_with(stream,
                                                       StreamState.firing)
        pm._run_pipeline.assert_called_once_with(stream, trigger_def,
                                    'test_fire_pipeline', pipeline_config)
        self.assertFalse(pm._error_stream.called)
        pm._complete_stream.assert_called_once_with(stream)
        self.assertTrue(ret)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_fire_stream_locked(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        stream = mock.MagicMock(name='stream')
        stream.name = 'test'
        pm.db = mock.MagicMock(spec=pm.db, name='db')
        pm.db.set_stream_state.side_effect = winch_db.LockError('locked!')
        trigger_def = mock.MagicMock(name='trigger_def')
        trigger_def.fire_pipeline = 'test_fire_pipeline'
        pm.trigger_map = dict(test=trigger_def)
        pipeline_config = mock.MagicMock(name='pipeline_config')
        pm.pipeline_config = dict(test_fire_pipeline=pipeline_config)
        pm._error_stream = mock.MagicMock(name='_error_stream')
        pm._complete_stream = mock.MagicMock(name='_complete_stream')
        pm._run_pipeline = mock.MagicMock(name='_run_pipeline')
        pm._run_pipeline.return_value = True

        ret = pm.fire_stream(stream)
        pm.db.set_stream_state.assert_called_once_with(stream, StreamState.firing)
        self.assertFalse(pm._run_pipeline.called)
        self.assertFalse(pm._error_stream.called)
        self.assertFalse(pm._complete_stream.called)
        self.assertFalse(ret)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_fire_stream_no_pipeline(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        stream = mock.MagicMock(name='stream')
        stream.name = 'test'
        pm.db = mock.MagicMock(spec=pm.db, name='db')
        pm.db.set_stream_state.return_value = stream
        trigger_def = mock.MagicMock(name='trigger_def')
        trigger_def.fire_pipeline = None
        pm.trigger_map = dict(test=trigger_def)
        pm._error_stream = mock.MagicMock(name='_error_stream')
        pm._complete_stream = mock.MagicMock(name='_complete_stream')
        pm._run_pipeline = mock.MagicMock(name='_run_pipeline')
        pm._run_pipeline.return_value = True

        ret = pm.fire_stream(stream)
        pm.db.set_stream_state.assert_called_once_with(stream,
                                                    StreamState.firing)
        self.assertFalse(pm._error_stream.called)
        self.assertFalse(pm._run_pipeline.called)
        pm._complete_stream.assert_called_once_with(stream)
        self.assertTrue(ret)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_fire_stream_error(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        stream = mock.MagicMock(name='stream')
        stream.name = 'test'
        pm.db = mock.MagicMock(spec=pm.db, name='db')
        pm.db.set_stream_state.return_value = stream
        trigger_def = mock.MagicMock(name='trigger_def')
        trigger_def.fire_pipeline = 'test_fire_pipeline'
        pm.trigger_map = dict(test=trigger_def)
        pipeline_config = mock.MagicMock(name='pipeline_config')
        pm.pipeline_config = dict(test_fire_pipeline=pipeline_config)
        pm._error_stream = mock.MagicMock(name='_error_stream')
        pm._complete_stream = mock.MagicMock(name='_complete_stream')
        pm._run_pipeline = mock.MagicMock(name='_run_pipeline')
        pm._run_pipeline.return_value = False

        ret = pm.fire_stream(stream)
        pm.db.set_stream_state.assert_called_once_with(stream,
                                                       StreamState.firing)
        pm._run_pipeline.assert_called_once_with(stream, trigger_def,
                                         'test_fire_pipeline',
                                         pipeline_config)
        self.assertFalse(pm._complete_stream.called)
        pm._error_stream.assert_called_once_with(stream)
        self.assertFalse(ret)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_expire_stream(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        stream = mock.MagicMock(name='stream')
        stream.name = 'test'
        pm.db = mock.MagicMock(spec=pm.db, name='db')
        pm.db.set_stream_state.return_value = stream
        trigger_def = mock.MagicMock(name='trigger_def')
        trigger_def.expire_pipeline = 'test_fire_pipeline'
        pm.trigger_map = dict(test=trigger_def)
        pipeline_config = mock.MagicMock(name='pipeline_config')
        pm.pipeline_config = dict(test_fire_pipeline=pipeline_config)
        pm._error_stream = mock.MagicMock(name='_error_stream')
        pm._complete_stream = mock.MagicMock(name='_complete_stream')
        pm._run_pipeline = mock.MagicMock(name='_run_pipeline')
        pm._run_pipeline.return_value = True

        ret = pm.expire_stream(stream)
        pm.db.set_stream_state.assert_called_once_with(stream,
                                                       StreamState.expiring)
        pm._run_pipeline.assert_called_once_with(stream, trigger_def,
                    'test_fire_pipeline', pipeline_config)
        self.assertFalse(pm._error_stream.called)
        pm._complete_stream.assert_called_once_with(stream)
        self.assertTrue(ret)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_expire_stream_locked(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        stream = mock.MagicMock(name='stream')
        stream.name = 'test'
        pm.db = mock.MagicMock(spec=pm.db, name='db')
        pm.db.set_stream_state.side_effect = winch_db.LockError('locked!')
        trigger_def = mock.MagicMock(name='trigger_def')
        trigger_def.expire_pipeline = 'test_fire_pipeline'
        pm.trigger_map = dict(test=trigger_def)
        pipeline_config = mock.MagicMock(name='pipeline_config')
        pm.pipeline_config = dict(test_fire_pipeline=pipeline_config)
        pm._expire_error_stream = mock.MagicMock(name='_expire_error_stream')
        pm._complete_stream = mock.MagicMock(name='_complete_stream')
        pm._run_pipeline = mock.MagicMock(name='_run_pipeline')
        pm._run_pipeline.return_value = True

        ret = pm.expire_stream(stream)
        pm.db.set_stream_state.assert_called_once_with(stream, StreamState.expiring)
        self.assertFalse(pm._run_pipeline.called)
        self.assertFalse(pm._expire_error_stream.called)
        self.assertFalse(pm._complete_stream.called)
        self.assertFalse(ret)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_expire_stream_no_pipeline(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        stream = mock.MagicMock(name='stream')
        stream.name = 'test'
        pm.db = mock.MagicMock(spec=pm.db, name='db')
        pm.db.set_stream_state.return_value = stream
        trigger_def = mock.MagicMock(name='trigger_def')
        trigger_def.expire_pipeline = None
        pm.trigger_map = dict(test=trigger_def)
        pm._expire_error_stream = mock.MagicMock(name='_expire_error_stream')
        pm._complete_stream = mock.MagicMock(name='_complete_stream')
        pm._run_pipeline = mock.MagicMock(name='_run_pipeline')
        pm._run_pipeline.return_value = True

        ret = pm.expire_stream(stream)
        pm.db.set_stream_state.assert_called_once_with(stream,
                                                    StreamState.expiring)
        self.assertFalse(pm._expire_error_stream.called)
        self.assertFalse(pm._run_pipeline.called)
        pm._complete_stream.assert_called_once_with(stream)
        self.assertTrue(ret)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_expire_stream_error(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        stream = mock.MagicMock(name='stream')
        stream.name = 'test'
        pm.db = mock.MagicMock(spec=pm.db, name='db')
        pm.db.set_stream_state.return_value = stream
        trigger_def = mock.MagicMock(name='trigger_def')
        trigger_def.expire_pipeline = 'test_fire_pipeline'
        pm.trigger_map = dict(test=trigger_def)
        pipeline_config = mock.MagicMock(name='pipeline_config')
        pm.pipeline_config = dict(test_fire_pipeline=pipeline_config)
        pm._expire_error_stream = mock.MagicMock(name='_expire_error_stream')
        pm._complete_stream = mock.MagicMock(name='_complete_stream')
        pm._run_pipeline = mock.MagicMock(name='_run_pipeline')
        pm._run_pipeline.return_value = False

        ret = pm.expire_stream(stream)
        pm.db.set_stream_state.assert_called_once_with(stream,
                                                       StreamState.expiring)
        pm._run_pipeline.assert_called_once_with(stream, trigger_def,
                                    'test_fire_pipeline', pipeline_config)
        self.assertFalse(pm._complete_stream.called)
        pm._expire_error_stream.assert_called_once_with(stream)
        self.assertFalse(ret)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_process_ready_streams_fire(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        pm.db = mock.MagicMock(spec=pm.db, name='db')
        stream = mock.MagicMock(name='stream')
        stream.name = "my_stream"
        tdef = mock.MagicMock(name='tdef')
        pm.trigger_map['my_stream'] = tdef
        pm.expire_stream = mock.MagicMock(name='expire_stream')
        pm.fire_stream = mock.MagicMock(name='fire_stream')
        pm.current_time = mock.MagicMock(name='current_time')
        pm.db.get_ready_streams.return_value = [stream]

        ret = pm.process_ready_streams(42)
        pm.db.get_ready_streams.assert_called_once_with(42,
                                pm.current_time.return_value, expire=False)
        pm.fire_stream.assert_called_once_with(stream)
        self.assertFalse(pm.expire_stream.called)
        self.assertEqual(ret, 1)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_process_ready_streams_expire(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        pm.db = mock.MagicMock(spec=pm.db, name='db')
        stream = mock.MagicMock(name='stream')
        stream.name = "my_stream"
        pm.expire_stream = mock.MagicMock(name='expire_stream')
        pm.fire_stream = mock.MagicMock(name='fire_stream')
        pm.current_time = mock.MagicMock(name='current_time')
        pm.db.get_ready_streams.return_value = [stream]

        ret = pm.process_ready_streams(42, expire=True)
        pm.db.get_ready_streams.assert_called_once_with(42,
                                pm.current_time.return_value, expire=True)
        pm.expire_stream.assert_called_once_with(stream)
        self.assertFalse(pm.fire_stream.called)
        self.assertEqual(ret, 1)

    @mock.patch.object(pipeline_manager.ConfigManager, 'wrap')
    def test_safe_get_debugger(self, mock_config_wrap):
        pm = pipeline_manager.PipelineManager('test')
        tdef = mock.MagicMock(name="tdef")
        tdef.name = "my trigger"
        tdef.debugger = self.debugger
        self.assertEqual(pm.safe_get_debugger(tdef), self.debugger)

        self.assertEqual(pm.safe_get_debugger(None)._name, "n/a")




import time
import logging
import random
import simport
import six

from winchester.db import DBInterface, DuplicateError, LockError
from winchester.config import ConfigManager, ConfigSection, ConfigItem
from winchester.definition import TriggerDefinition
from winchester.models import StreamState
from winchester import time_sync as ts
from winchester.trigger_manager import TriggerManager


logger = logging.getLogger(__name__)


class PipelineError(Exception):
    pass


class PipelineExecutionError(PipelineError):
    def __init__(self, msg="", cause=None):
        super(PipelineExecutionError, self).__init__("%s: caused by %s" % (msg, repr(cause)))
        self.cause = cause


class PipelineConfigError(PipelineError):
    pass


class Pipeline(object):

    @classmethod
    def check_handler_config(cls, conf, handler_map):
        if isinstance(conf, six.string_types):
            conf = dict(name=conf, params=dict())
        if 'name' not in conf:
            raise PipelineConfigError("Handler name not in config! %s" % str(conf))
        if 'params' not in conf:
            conf['params'] = {}
        if conf['name'] not in handler_map:
            raise PipelineConfigError("Unknown handler in pipeline config %s" % conf['name'])
        return conf

    def __init__(self, name, config, handler_map):
        self.name = name
        self.handlers = []
        self.env = dict()
        for handler_conf in config:
            name = handler_conf['name']
            params = handler_conf['params']
            handler_class = handler_map[name]
            try:
                handler = handler_class(**params)
            except Exception as e:
                logger.exception("Error initalizing handler %s for pipeline %s" %
                                 (handler_class, self.name))
                raise PipelineExecutionError("Error loading pipeline", e)
            self.handlers.append(handler)

    def handle_events(self, events, stream, debugger):
        self.env['stream_id'] = stream.id
        event_ids = set(e['message_id'] for e in events)
        try:
            for handler in self.handlers:
                events = handler.handle_events(events, self.env)
            debugger.bump_counter("Pre-commit successful")
        except Exception as e:
            logger.exception("Error processing pipeline %s" % self.name)
            debugger.bump_counter("Pipeline error")
            self.rollback(debugger)
            raise PipelineExecutionError("Error in pipeline", e)
        new_events = [e for e in events if e['message_id'] not in event_ids]
        self.commit(debugger)
        return new_events

    def commit(self, debugger):
        for handler in self.handlers:
            try:
                handler.commit()
                debugger.bump_counter("Commit successful")
            except:
                debugger.bump_counter("Commit error")
                logger.exception("Commit error on handler in pipeline %s" % self.name)

    def rollback(self, debugger):
        for handler in self.handlers:
            try:
                handler.rollback()
                debugger.bump_counter("Rollback successful")
            except:
                debugger.bump_counter("Rollback error")
                logger.exception("Rollback error on handler in pipeline %s" % self.name)


class PipelineManager(object):

    @classmethod
    def config_description(cls):
        configs = TriggerManager.config_description()
        configs.update(dict(
                    pipeline_handlers=ConfigItem(required=True,
                                 help="dictionary of pipeline handlers to load "
                                       "Classes specified with simport syntax. "
                                       "simport docs for more info"),
                    pipeline_worker_batch_size=ConfigItem(
                                 help="Number of streams for pipeline "
                                      "worker(s) to load at a time",
                                      default=1000),
                    pipeline_worker_delay=ConfigItem(
                                 help="Number of seconds for pipeline worker "
                                      "to sleep when it finds no streams to "
                                      "process", default=10),
                    pipeline_config=ConfigItem(required=True,
                                       help="Name of pipeline config file "
                                            "defining the handlers for each "
                                            "pipeline."),
                    purge_completed_streams=ConfigItem(
                                       help="Delete successfully proccessed "
                                            "streams when finished?",
                                            default=True),
                   ))
        return configs

    def __init__(self, config, db=None, pipeline_handlers=None,
                 pipeline_config=None, trigger_defs=None, time_sync=None,
                 proc_name='pipeline_worker'):
        #name used to distinguish worker processes in logs
        self.proc_name = proc_name

        logger.debug("PipelineManager(%s): Using config: %s" % (self.proc_name, str(config)))
        config = ConfigManager.wrap(config, self.config_description())
        self.config = config
        config.check_config()
        config.add_config_path(*config['config_path'])
        if time_sync is None:
            time_sync = ts.TimeSync()
        self.time_sync = time_sync

        if db is not None:
            self.db = db
        else:
            self.db = DBInterface(config['database'])

        if pipeline_handlers is not None:
            self.pipeline_handlers = pipeline_handlers
        else:
            self.pipeline_handlers = self._load_plugins(config['pipeline_handlers'])
        logger.debug("Pipeline handlers: %s" % str(self.pipeline_handlers))

        if pipeline_config is not None:
            self.pipeline_config = pipeline_config
        else:
            self.pipeline_config = config.load_file(config['pipeline_config'])

        logger.debug("Pipeline config: %s" % str(self.pipeline_config))
        for pipeline, handler_configs in self.pipeline_config.items():
            self.pipeline_config[pipeline] = [Pipeline.check_handler_config(conf,
                                                self.pipeline_handlers)
                                              for conf in handler_configs]

        if trigger_defs is not None:
            self.trigger_definitions = trigger_defs
        else:
            defs = config.load_file(config['trigger_definitions'])
            logger.debug("Loaded trigger definitions %s" % str(defs))
            self.trigger_definitions = [TriggerDefinition(conf, None) for conf in defs]
        self.trigger_map = dict((tdef.name, tdef) for tdef in self.trigger_definitions)

        self.trigger_manager = TriggerManager(self.config, db=self.db,
                                              trigger_defs=self.trigger_definitions,
                                              time_sync=time_sync)

        self.pipeline_worker_batch_size = config['pipeline_worker_batch_size']
        self.pipeline_worker_delay = config['pipeline_worker_delay']
        self.statistics_period = config['statistics_period']
        self.purge_completed_streams = config['purge_completed_streams']
        self.streams_fired = 0
        self.streams_expired = 0
        self.streams_loaded = 0
        self.last_status = self.current_time()

    @classmethod
    def _load_plugins(cls, plug_map, defaults=None):
        plugins = dict()
        if defaults is not None:
            plugins.update(defaults)
        for name, cls_string in plug_map.items():
            try:
                plugins[name] = simport.load(cls_string)
            except simport.ImportFailed as e:
                logger.error("Could not load plugin %s: Import failed. %s" % (
                             name, e))
            except (simport.MissingMethodOrFunction,
                    simport.MissingModule,
                    simport.BadDirectory) as e:
                log.error("Could not load plugin %s: Not found. %s" % (
                          name, e))
        return plugins

    def current_time(self):
        # here so it's easily overridden.
        return self.time_sync.current_time()

    def _log_statistics(self):
        logger.info("Loaded %s streams. Fired %s, Expired %s." % (
                    self.streams_loaded, self.streams_fired, self.streams_expired))
        self.streams_fired = 0
        self.streams_expired = 0
        self.streams_loaded = 0
        self.last_status = self.current_time()

        self.trigger_manager.debug_manager.dump_debuggers()

    def add_new_events(self, events):
        for event in events:
            self.trigger_manager.add_event(event)

    def _run_pipeline(self, stream, trigger_def, pipeline_name,
                      pipeline_config):
        events = self.db.get_stream_events(stream)
        debugger = trigger_def.debugger
        try:
            pipeline = Pipeline(pipeline_name, pipeline_config, self.pipeline_handlers)
            new_events = pipeline.handle_events(events, stream, debugger)
        except PipelineExecutionError:
            logger.error("Exception in pipeline %s handling stream %s" % (
                          pipeline_name, stream.id))
            return False
        if new_events:
            self.add_new_events(new_events)
        return True

    def _complete_stream(self, stream):
        if self.purge_completed_streams:
            self.db.purge_stream(stream)
        else:
            try:
                self.db.set_stream_state(stream, StreamState.completed)
            except LockError:
                logger.error("Stream %s locked while trying to set 'complete' state! "
                             "This should not happen." % stream.id)

    def _error_stream(self, stream):
        try:
            self.db.set_stream_state(stream, StreamState.error)
        except LockError:
            logger.error("Stream %s locked while trying to set 'error' state! "
                         "This should not happen." % stream.id)

    def _expire_error_stream(self, stream):
        try:
            self.db.set_stream_state(stream, StreamState.expire_error)
        except LockError:
            logger.error("Stream %s locked while trying to set 'expire_error' state! "
                         "This should not happen." % stream.id)

    def safe_get_debugger(self, trigger_def):
        return trigger_def.debugger if trigger_def is not None else \
            self.trigger_manager.debug_manager.get_debugger(None)

    def fire_stream(self, stream):
        trigger_def = self.trigger_map.get(stream.name)
        debugger = self.safe_get_debugger(trigger_def)
        try:
            stream = self.db.set_stream_state(stream, StreamState.firing)
        except LockError:
            logger.debug("Stream %s locked. Moving on..." % stream.id)
            debugger.bump_counter("Locked")
            return False
        logger.debug("Firing Stream %s." % stream.id)
        if trigger_def is None:
            debugger.bump_counter("Unknown trigger def '%s'" % stream.name)
            logger.error("Stream %s has unknown trigger definition %s" % (
                         stream.id, stream.name))
            self._error_stream(stream)
            return False
        pipeline = trigger_def.fire_pipeline
        if pipeline is not None:
            pipe_config = self.pipeline_config.get(pipeline)
            if pipe_config is None:
                debugger.bump_counter("Unknown pipeline '%s'" % pipeline)
                logger.error("Trigger %s for stream %s has unknown "
                             "pipeline %s" % (stream.name, stream.id,
                                              pipeline))
                self._error_stream(stream)
            if not self._run_pipeline(stream, trigger_def, pipeline,
                                      pipe_config):
                self._error_stream(stream)
                return False
        else:
            logger.debug("No fire pipeline for stream %s. Nothing to do." % (
                         stream.id))
            debugger.bump_counter("No fire pipeline for '%s'" % stream.name)
        self._complete_stream(stream)
        debugger.bump_counter("Streams fired")
        self.streams_fired +=1
        return True

    def expire_stream(self, stream):
        trigger_def = self.trigger_map.get(stream.name)
        debugger = self.safe_get_debugger(trigger_def)
        try:
            stream = self.db.set_stream_state(stream, StreamState.expiring)
        except LockError:
            debugger.bump_counter("Locked")
            logger.debug("Stream %s locked. Moving on..." % stream.id)
            return False
        logger.debug("Expiring Stream %s." % stream.id)
        if trigger_def is None:
            debugger.bump_counter("Unknown trigger def '%s'" % stream.name)
            logger.error("Stream %s has unknown trigger definition %s" % (
                         stream.id, stream.name))
            self._expire_error_stream(stream)
            return False
        pipeline = trigger_def.expire_pipeline
        if pipeline is not None:
            pipe_config = self.pipeline_config.get(pipeline)
            if pipe_config is None:
                debugger.bump_counter("Unknown pipeline '%s'" % pipeline)
                logger.error("Trigger %s for stream %s has unknown pipeline %s" % (
                            stream.name, stream.id, pipeline))
                self._expire_error_stream(stream)
            if not self._run_pipeline(stream, trigger_def, pipeline,
                                      pipe_config):
                self._expire_error_stream(stream)
                return False
        else:
            logger.debug("No expire pipeline for stream %s. Nothing to do." % (
                         stream.id))
            debugger.bump_counter("No expire pipeline for '%s'" % stream.name)
        self._complete_stream(stream)
        debugger.bump_counter("Streams expired")
        self.streams_expired +=1
        return True

    def process_ready_streams(self, batch_size, expire=False):
        streams = self.db.get_ready_streams(batch_size, self.current_time(),
                                            expire=expire)
        stream_ct = len(streams)
        if expire:
            logger.debug("Loaded %s streams to expire." % stream_ct)
        else:
            logger.debug("Loaded %s streams to fire." % stream_ct)

        random.shuffle(streams)
        for stream in streams:
            if expire:
                self.expire_stream(stream)
            else:
                self.fire_stream(stream)
        self.streams_loaded += stream_ct
        return stream_ct

    def run(self):
        while True:
            fire_ct = self.process_ready_streams(self.pipeline_worker_batch_size)
            expire_ct = self.process_ready_streams(self.pipeline_worker_batch_size,
                                                   expire=True)

            if (self.current_time() - self.last_status).seconds > self.statistics_period:
                self._log_statistics()

            if not fire_ct and not expire_ct:
                logger.debug("No streams to fire or expire. Sleeping...")
                time.sleep(self.pipeline_worker_delay)



#for Python2.6 compatability.
import unittest2 as unittest

import mock
import logging

import datetime
import timex
from winchester import db, models

logging.basicConfig()

TEST_DATA = [
    {'event_type': [
        dict(id=1, desc='test.thing.begin'),
        dict(id=2, desc='test.thing.end'),
        dict(id=3, desc='test.otherthing.foo'),
        ]},
    {'event': [
        dict(id=1,
         message_id='1234-5678-001',
         generated=datetime.datetime(2014,8,1,10,20,45,453201),
         event_type_id=1,),
        dict(id=2,
         message_id='1234-5678-002',
         generated=datetime.datetime(2014,8,1,15,25,45,453201),
         event_type_id=2,),
        dict(id=3,
         message_id='1234-5678-003',
         generated=datetime.datetime(2014,8,1,2,10,12,0),
         event_type_id=3,),
        dict(id=4,
         message_id='1234-5678-004',
         generated=datetime.datetime(2014,8,1,4,57,55,42),
         event_type_id=3,),
        ]},
    {'trait': [
        dict(event_id=1, name='instance_id', type=int(models.Datatype.string),
         t_string='aaaa-bbbb-cccc-dddd'),
        dict(event_id=1, name='memory_mb', type=int(models.Datatype.int),
         t_int=1024),
        dict(event_id=1, name='test_weight', type=int(models.Datatype.float),
         t_float=20112.42),
        dict(event_id=1, name='launched_at', type=int(models.Datatype.datetime),
         t_datetime=datetime.datetime(2014,7,1,2,30,45,453201)),
        ]},
    {'stream': [
        dict(id=1, first_event=datetime.datetime(2014,8,1,2,10,12,0),
         last_event=datetime.datetime(2014,8,1,4,57,55,42),
         name='test_trigger',
         expire_timestamp=datetime.datetime(2014,8,2,4,57,55,42),
         state=int(models.StreamState.active),
         state_serial_no=0),
        dict(id=2, first_event=datetime.datetime(2014,8,1,15,25,45,453201),
         last_event=datetime.datetime(2014,8,1,15,25,45,453201),
         name='another_test_trigger',
         expire_timestamp=datetime.datetime(2014,8,2,4,57,55,42),
         state=int(models.StreamState.active),
         state_serial_no=0),
        dict(id=3, first_event=datetime.datetime(2014,8,1,15,25,45,453201),
         last_event=datetime.datetime(2014,8,1,15,25,45,453201),
         name='fire_test_trigger',
         fire_timestamp=datetime.datetime(2014,8,10,6,0,0,42),
         expire_timestamp=datetime.datetime(2014,8,15,6,0,0,42),
         state=int(models.StreamState.active),
         state_serial_no=0),
        dict(id=4, first_event=datetime.datetime(2014,8,1,15,25,45,453201),
         last_event=datetime.datetime(2014,8,1,15,25,45,453201),
         name='fire_test_trigger',
         fire_timestamp=datetime.datetime(2014,8,11,6,0,0,42),
         expire_timestamp=datetime.datetime(2014,8,16,0,0,0,42),
         state=int(models.StreamState.active),
         state_serial_no=0),
        dict(id=5, first_event=datetime.datetime(2014,8,1,15,25,45,453201),
         last_event=datetime.datetime(2014,8,1,15,25,45,453201),
         name='reset_test_trigger',
         fire_timestamp=datetime.datetime(2014,8,11,6,0,0,42),
         expire_timestamp=datetime.datetime(2014,8,16,0,0,0,42),
         state=int(models.StreamState.error),
         state_serial_no=0),
        dict(id=6, first_event=datetime.datetime(2014,8,1,15,25,45,453201),
         last_event=datetime.datetime(2014,8,1,15,25,45,453201),
         name='reset_test_trigger',
         fire_timestamp=datetime.datetime(2014,8,11,6,0,0,42),
         expire_timestamp=datetime.datetime(2014,8,16,0,0,0,42),
         state=int(models.StreamState.expire_error),
         state_serial_no=0),
        dict(id=7, first_event=datetime.datetime(2014,8,1,15,25,45,453201),
         last_event=datetime.datetime(2014,8,1,15,25,45,453201),
         name='reset_test_trigger',
         fire_timestamp=datetime.datetime(2014,8,11,6,0,0,42),
         expire_timestamp=datetime.datetime(2014,8,16,0,0,0,42),
         state=int(models.StreamState.retry_fire),
         state_serial_no=0),
        dict(id=8, first_event=datetime.datetime(2014,8,1,15,25,45,453201),
         last_event=datetime.datetime(2014,8,1,15,25,45,453201),
         name='reset_test_trigger',
         fire_timestamp=datetime.datetime(2014,8,11,6,0,0,42),
         expire_timestamp=datetime.datetime(2014,8,16,0,0,0,42),
         state=int(models.StreamState.retry_expire),
         state_serial_no=0),
        ]},
    {'streamevent': [
        dict(stream_id=1, event_id=3),
        dict(stream_id=1, event_id=4),
        dict(stream_id=2, event_id=2),
        dict(stream_id=3, event_id=2),
        dict(stream_id=3, event_id=1),
        dict(stream_id=4, event_id=2),
        ]},
    {'dist_trait': [
        dict(stream_id=1, name='instance_id', type=int(models.Datatype.string),
         dt_string='zzzz-xxxx-yyyy-wwww'),
        dict(stream_id=1, name='memory_mb', type=int(models.Datatype.int),
         dt_int=4096),
        dict(stream_id=1, name='test_weight', type=int(models.Datatype.float),
         dt_float=3.1415),
        dict(stream_id=1, name='launched_at', type=int(models.Datatype.datetime),
         dt_datetime=datetime.datetime(2014,7,8,9,40,50,77777)),
        dict(stream_id=1, name='timestamp', type=int(models.Datatype.timerange),
         dt_timerange_begin=datetime.datetime(2014,7,8,0,0,0,27),
         dt_timerange_end=datetime.datetime(2014,7,9,0,0,0,27)),
        ]},
]


def create_tables(dbi):
    #used for tests
    models.Base.metadata.create_all(dbi.engine)


def load_fixture_data(dbi, data):
    #Used for tests. This is fugly, refactor later (mdragon)
    for table in data:
        for table_name, rows in table.items():
            for row in rows:
                cols = []
                vals = []
                for col, val in row.items():
                    cols.append(col)
                    vals.append(val)

                q = ("INSERT into %(table)s (%(colnames)s) VALUES (%(qs)s)" %
                     dict(table=table_name,
                          colnames=','.join(cols),
                          qs=','.join(('?',) * len(vals)),))
                dbi.engine.execute(q, vals)


class TestDB(unittest.TestCase):

    def setUp(self):
        super(TestDB, self).setUp()
        self.db = db.DBInterface(dict(url='sqlite://'))
        create_tables(self.db)
        load_fixture_data(self.db, TEST_DATA)
        logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)

        self.events = {}
        for event in TEST_DATA[1]['event']:
            self.events[event['message_id']] = event

    def tearDown(self):
        logging.getLogger('sqlalchemy.engine').setLevel(logging.WARNING)
        self.db.close()

    def test_get_event_type(self):
        t = self.db.get_event_type('test.thing.begin')
        self.assertEqual(t.id, 1)
        t = self.db.get_event_type('test.not_in_db')
        self.assertEqual(t.id, 4) #next unused id.

    def test_create_event(self):
        message_id = '9876-0001-0001'
        event_type = 'test.thing.begin'
        timestamp = datetime.datetime(2014,7,4,12,7,21,4096)
        traits = dict(test_string='foobar',
                      test_number=42,
                      test_float=3.1415,
                      test_date=datetime.datetime(2014,7,1,0,0,0,0),
                      somevalue=u'A fine test string')
        self.db.create_event(message_id, event_type, timestamp, traits)
        event = self.db.get_event_by_message_id(message_id)
        self.assertEqual(len(event), 8)
        self.assertEqual(event['message_id'], message_id)
        self.assertEqual(event['event_type'], event_type)
        self.assertEqual(event['timestamp'], timestamp)
        for name, value in traits.items():
            self.assertEqual(event[name], value)
            if type(value) == str:
                t_value = unicode
            else:
                t_value = type(value)
            self.assertEqual(type(event[name]), t_value)

    def test_create_event_duplicate(self):
        message_id = '9876-0001-0001'
        event_type = 'test.thing.begin'
        timestamp = datetime.datetime(2014,7,4,12,7,21,4096)
        traits = dict(test_string='foobar',
                      test_number=42,
                      test_float=3.1415,
                      test_date=datetime.datetime(2014,7,1,0,0,0,0),
                      somevalue=u'A fine test string')
        self.db.create_event(message_id, event_type, timestamp, traits)
        with self.assertRaises(db.DuplicateError):
            self.db.create_event(message_id, event_type, timestamp, traits)

    def test_get_event_by_message_id(self):
        event = self.db.get_event_by_message_id('1234-5678-001')
        self.assertEqual(len(event), 7)
        expected = dict(message_id='1234-5678-001',
                        event_type='test.thing.begin',
                        timestamp=datetime.datetime(2014,8,1,10,20,45,453201),
                        instance_id='aaaa-bbbb-cccc-dddd',
                        memory_mb=1024,
                        test_weight=20112.42,
                        launched_at=datetime.datetime(2014,7,1,2,30,45,453201),)
        self.assertDictContainsSubset(expected, event)

    def test_get_stream_events(self):
        stream = self.db.get_stream_by_id(1)
        events =  self.db.get_stream_events(stream)
        self.assertEqual(len(events), 2)
        self.assertIn('1234-5678-003', [e['message_id'] for e in events])
        self.assertIn('1234-5678-004', [e['message_id'] for e in events])

    def test_create_stream(self):
        event = dict(message_id='1234-5678-001',
                     event_type='test.thing.begin',
                     timestamp=datetime.datetime(2014,8,1,10,20,45,453201),
                     instance_id='aaaa-bbbb-cccc-dddd',
                     memory_mb=1024,
                     test_weight=20112.42,
                     launched_at=datetime.datetime(2014,7,1,2,30,45,453201),)
        timestamp = timex.TimeRange(datetime.datetime(2014,8,1,0,0,0,27),
                                    datetime.datetime(2014,2,2,0,0,0,27))
        dist_traits = dict(timestamp=timestamp,
                           instance_id='aaaa-bbbb-cccc-dddd')

        class MockTimestamp(object):
            pass

        mock_expire_value = datetime.datetime(2014,8,2,12,12,12,12)

        def mock_time_expr(first, last):
            self.assertEqual(first, datetime.datetime(2014,8,1,10,20,45,453201))
            self.assertEqual(last, datetime.datetime(2014,8,1,10,20,45,453201))
            t = MockTimestamp()
            t.timestamp = mock_expire_value
            return t

        stream = self.db.create_stream('test_create_stream', event, dist_traits, mock_time_expr)
        self.assertEqual(stream.name, 'test_create_stream')
        self.assertEqual(stream.first_event, datetime.datetime(2014,8,1,10,20,45,453201))
        self.assertEqual(stream.last_event, datetime.datetime(2014,8,1,10,20,45,453201))
        self.assertEqual(stream.expire_timestamp, mock_expire_value)
        self.assertIsNone(stream.fire_timestamp)
        self.assertEqual(stream.state, models.StreamState.active)
        self.assertEqual(stream.state_serial_no, 0)
        self.assertTrue(self.db.stream_has_dist_trait(stream.id, 'timestamp', timestamp))
        self.assertTrue(self.db.stream_has_dist_trait(stream.id,
                                                      'instance_id',
                                                      'aaaa-bbbb-cccc-dddd'))
        events = self.db.get_stream_events(stream)
        self.assertEqual(len(events), 1)
        self.assertEqual(events[0]['message_id'], '1234-5678-001')

    def test_add_event_stream(self):
        stream = self.db.get_stream_by_id(1)
        event = dict(message_id='1234-5678-001',
                     event_type='test.thing.begin',
                     timestamp=datetime.datetime(2014,8,1,10,20,45,453201),
                     instance_id='aaaa-bbbb-cccc-dddd',
                     memory_mb=1024,
                     test_weight=20112.42,
                     launched_at=datetime.datetime(2014,7,1,2,30,45,453201),)

        class MockTimestamp(object):
            pass

        mock_expire_value = datetime.datetime(2014,8,2,12,12,12,12)

        def mock_time_expr(first, last):
            self.assertEqual(first, datetime.datetime(2014,8,1,2,10,12,0))
            self.assertEqual(last, datetime.datetime(2014,8,1,10,20,45,453201))
            t = MockTimestamp()
            t.timestamp = mock_expire_value
            return t

        self.db.add_event_stream(stream, event, mock_time_expr)
        self.assertEqual(stream.expire_timestamp, mock_expire_value)
        self.assertEqual(stream.first_event, datetime.datetime(2014,8,1,2,10,12,0))
        self.assertEqual(stream.last_event, datetime.datetime(2014,8,1,10,20,45,453201))
        events = self.db.get_stream_events(stream)
        self.assertEqual(len(events), 3)
        self.assertIn('1234-5678-001', [e['message_id'] for e in events])
        self.assertIn('1234-5678-003', [e['message_id'] for e in events])
        self.assertIn('1234-5678-004', [e['message_id'] for e in events])

    def test_stream_dist_traits(self):
        with self.db.in_session() as session:
            stream = self.db.get_stream_by_id(1, session=session)
            dist_traits = stream.distinguished_by_dict
        self.assertEqual(len(dist_traits), 5)
        self.assertIn('instance_id', dist_traits)
        self.assertEqual(dist_traits['instance_id'], 'zzzz-xxxx-yyyy-wwww')
        self.assertEqual(type(dist_traits['instance_id']), unicode)
        self.assertIn('memory_mb', dist_traits)
        self.assertEqual(dist_traits['memory_mb'], 4096)
        self.assertEqual(type(dist_traits['memory_mb']), int)
        self.assertIn('test_weight', dist_traits)
        self.assertEqual(dist_traits['test_weight'], 3.1415)
        self.assertEqual(type(dist_traits['test_weight']), float)
        self.assertIn('launched_at', dist_traits)
        self.assertEqual(dist_traits['launched_at'], datetime.datetime(2014,7,8,9,40,50,77777))
        self.assertEqual(type(dist_traits['launched_at']), datetime.datetime)
        self.assertIn('timestamp', dist_traits)
        timestamp = dist_traits['timestamp']
        self.assertEqual(type(timestamp), timex.TimeRange)
        self.assertEqual(timestamp.begin, datetime.datetime(2014,7,8,0,0,0,27))
        self.assertEqual(timestamp.end, datetime.datetime(2014,7,9,0,0,0,27))

    def test_stream_has_dist_trait(self):
        #this mostly tests that the polymorphic trait comparisons are working.
        dt = self.db.stream_has_dist_trait(1, 'instance_id', 'zzzz-xxxx-yyyy-wwww')
        self.assertIsNotNone(dt)
        self.assertEqual(len(dt), 1)
        self.assertIn('instance_id', dt)
        self.assertEqual(dt['instance_id'], 'zzzz-xxxx-yyyy-wwww')

        dt = self.db.stream_has_dist_trait(1, 'memory_mb', 4096)
        self.assertIsNotNone(dt)
        self.assertEqual(len(dt), 1)
        self.assertIn('memory_mb', dt)
        self.assertEqual(dt['memory_mb'], 4096)

        dt = self.db.stream_has_dist_trait(1, 'test_weight', 3.1415)
        self.assertIsNotNone(dt)
        self.assertEqual(len(dt), 1)
        self.assertIn('test_weight', dt)
        self.assertEqual(dt['test_weight'], 3.1415)

        launched = datetime.datetime(2014,7,8,9,40,50,77777)
        dt = self.db.stream_has_dist_trait(1, 'launched_at', launched)
        self.assertIsNotNone(dt)
        self.assertEqual(len(dt), 1)
        self.assertIn('launched_at', dt)
        self.assertEqual(dt['launched_at'],  launched)

        timestamp = timex.TimeRange(datetime.datetime(2014,7,8,0,0,0,27),
                                    datetime.datetime(2014,7,9,0,0,0,27))
        dt = self.db.stream_has_dist_trait(1, 'timestamp', timestamp)
        self.assertIsNotNone(dt)
        self.assertEqual(len(dt), 1)
        self.assertIn('timestamp', dt)
        self.assertEqual(dt['timestamp'].begin, timestamp.begin)
        self.assertEqual(dt['timestamp'].end, timestamp.end)

    def test_get_active_stream(self):
        timestamp = timex.TimeRange(datetime.datetime(2014,7,8,0,0,0,27),
                                    datetime.datetime(2014,7,9,0,0,0,27))
        dist_traits = dict(instance_id='zzzz-xxxx-yyyy-wwww',
                           memory_mb=4096,
                           test_weight=3.1415,
                           launched_at=datetime.datetime(2014,7,8,9,40,50,77777),
                           timestamp=timestamp)
        current_time = datetime.datetime(2014,8,2,1,0,0,02)
        stream = self.db.get_active_stream('test_trigger', dist_traits, current_time)
        self.assertIsNotNone(stream)
        self.assertEqual(stream.id, 1)
        current_time = datetime.datetime(2014,8,3,1,0,0,02)
        stream = self.db.get_active_stream('test_trigger', dist_traits, current_time)
        self.assertIsNone(stream)

    def test_stream_ready_to_fire(self):
        stream = self.db.get_stream_by_id(1)
        fire_time = datetime.datetime(2014, 8, 2, 12, 21, 2, 2)
        self.db.stream_ready_to_fire(stream, fire_time)
        stream = self.db.get_stream_by_id(1)
        self.assertEqual(stream.fire_timestamp, fire_time)

    def test_get_ready_streams_fire(self):
        current_time = datetime.datetime(2014,8,12,0,0,0,42)
        streams = self.db.get_ready_streams(10, current_time)
        self.assertEqual(len(streams), 3)
        stream_ids = [stream.id for stream in streams]
        self.assertIn(3, stream_ids)
        self.assertIn(4, stream_ids)
        self.assertIn(7, stream_ids)

        current_time = datetime.datetime(2014,8,10,12,0,0,42)
        streams = self.db.get_ready_streams(10, current_time)
        self.assertEqual(len(streams), 1)
        stream_ids = [stream.id for stream in streams]
        self.assertIn(3, stream_ids)

        current_time = datetime.datetime(2014,8,12,0,0,0,42)
        streams = self.db.get_ready_streams(1, current_time)
        self.assertEqual(len(streams), 1)

    def test_get_ready_streams_expire(self):
        current_time = datetime.datetime(2014,8,17,0,0,0,42)
        streams = self.db.get_ready_streams(10, current_time, expire=True)
        self.assertEqual(len(streams), 5)
        stream_ids = [stream.id for stream in streams]
        self.assertIn(1, stream_ids)
        self.assertIn(2, stream_ids)
        self.assertIn(3, stream_ids)
        self.assertIn(4, stream_ids)
        self.assertIn(8, stream_ids)

        current_time = datetime.datetime(2014,8,10,12,0,0,42)
        streams = self.db.get_ready_streams(10, current_time, expire=True)
        self.assertEqual(len(streams), 2)
        stream_ids = [stream.id for stream in streams]
        self.assertIn(1, stream_ids)
        self.assertIn(2, stream_ids)

        current_time = datetime.datetime(2014,8,17,0,0,0,42)
        streams = self.db.get_ready_streams(1, current_time, expire=True)
        self.assertEqual(len(streams), 1)

    def test_set_stream_state_sucess(self):
        stream = self.db.get_stream_by_id(1)
        old_serial = stream.state_serial_no
        new_stream = self.db.set_stream_state(stream, models.StreamState.firing)
        self.assertEqual(new_stream.state, models.StreamState.firing)
        self.assertEqual(new_stream.state_serial_no, old_serial + 1)

    def test_set_stream_state_locked(self):
        stream = self.db.get_stream_by_id(1)
        self.db.set_stream_state(stream, models.StreamState.firing)
        with self.assertRaises(db.LockError):
            self.db.set_stream_state(stream, models.StreamState.firing)

    def test_reset_stream_fire(self):
        stream = self.db.get_stream_by_id(5)
        stream = self.db.reset_stream(stream)
        self.assertEqual(stream.state, models.StreamState.retry_fire)

    def test_reset_stream_expire(self):
        stream = self.db.get_stream_by_id(6)
        stream = self.db.reset_stream(stream)
        self.assertEqual(stream.state, models.StreamState.retry_expire)

    def test_purge_stream(self):
        stream = self.db.get_stream_by_id(1)
        self.db.purge_stream(stream)
        with self.assertRaises(db.NoSuchStreamError):
            self.db.get_stream_by_id(1)

    def test_find_stream_count(self):
        count = self.db.find_streams(count=True)
        self.assertEqual([{'count': 8}], count)

    def test_find_stream_limit(self):
        streams = self.db.find_streams(limit=2)
        self.assertEqual(len(streams), 2)
        self.assertEqual(streams[0]['id'], 8)
        self.assertEqual(streams[1]['id'], 7)

    def test_find_stream_limit_asc(self):
        streams = self.db.find_streams(limit=2, mark='+')
        self.assertEqual(len(streams), 2)
        self.assertEqual(streams[0]['id'], 1)
        self.assertEqual(streams[1]['id'], 2)

    def test_find_stream_mark(self):
        streams = self.db.find_streams(mark='7')
        for stream in streams:
            self.assertIn('_mark', stream)
        self.assertEqual(streams[0]['id'], 6)
        self.assertEqual(streams[1]['id'], 5)

    def test_find_stream_mark_asc(self):
        streams = self.db.find_streams(mark='+2')
        for stream in streams:
            self.assertIn('_mark', stream)
        self.assertEqual(streams[0]['id'], 3)
        self.assertEqual(streams[1]['id'], 4)

    def test_find_events(self):
        events = self.db.find_events()
        self.assertEqual(4, len(events))
        for event in events:
            self.assertTrue(event['message_id'] in self.events)

    def test_find_event_count(self):
        count = self.db.find_events(count=True)
        self.assertEqual([{'count': 4}], count)

    def test_find_events_date_filter(self):
        _from = datetime.datetime(2014,8,1,10)
        _to = datetime.datetime(2014,8,1,16)
        events = self.db.find_events(from_datetime=_from, to_datetime=_to)
        self.assertEqual(2, len(events))
        msg_ids = [event['message_id'] for event in events]
        for good in ['1234-5678-001', '1234-5678-002']:
            self.assertTrue(good in msg_ids)

    def test_find_events_event_type(self):
        events = self.db.find_events(event_name='test.otherthing.foo')
        self.assertEqual(2, len(events))
        for event in events:
            self.assertTrue(event['event_type'], 'test.otherthing.foo')

    def test_find_events_traits(self):
        traits = {'memory_mb': 1024}
        events = self.db.find_events(traits=traits)
        self.assertEqual(1, len(events))
        self.assertTrue(events[0]['message_id'], '1234-5678-001')

    def test_find_events_limit(self):
        events = self.db.find_events(limit=2)
        self.assertEqual(2, len(events))

    def test_find_events_mark(self):
        events = self.db.find_events(limit=2)
        self.assertEqual(2, len(events))
        msg_ids = [event['message_id'] for event in events]

        events = self.db.find_events(limit=2, mark=events[-1]['_mark'])
        self.assertEqual(2, len(events))
        for event in events:
            self.assertTrue(event['message_id'] not in msg_ids)
